#include "config.h"
#include "particledata.h"
#include "sph.h"
#include "init.h"
#include "utils.h"

#include <vector>
#include <stdexcept>
#include <cmath>
#include <cstdlib>
#include <iostream>


using std::vector; 
using std::domain_error; 
using std::cout;
using std::endl; 

#ifdef ONE_D_SHOCKTUBE
void init_shocktube(vector<ParticleData>& sph); 
#endif
#ifdef TWO_D_KH
void init_kh(vector<ParticleData>& sph); 
#endif
#ifdef THREE_D_SEDOV
void init_sedov(vector<ParticleData>& sph); 
#endif

void initialize(vector<ParticleData>& sph) {

#ifdef ONE_D_SHOCKTUBE
  init_shocktube(sph); 
#endif

#ifdef TWO_D_KH
  init_kh(sph); 
#endif

#ifdef THREE_D_SEDOV
  init_sedov(sph); 
#endif

}

#ifdef ONE_D_SHOCKTUBE
void init_shocktube(vector<ParticleData>& sph)
{
  sph.resize(Ngrid); 
  sph.clear(); 

  typedef vector<ParticleData>::size_type p_sz; 
  const real rho_l = 1.0; 
  const real rho_r = 0.25; 
  const real p_l = 1.0; 
  const real p_r = 0.1; 
  const real v_l = 0; 
  const real v_r = 0; 

  const real mgas = 0.5 * (rho_l + rho_r)/static_cast<double>(Ngrid); 
  const p_sz n_l = Ngrid/(1.e0 + rho_r/rho_l); 
  const p_sz n_r = Ngrid - n_l; 
  const double dx_l = 0.5/static_cast<double>(n_l); 
  const double dx_r = 0.5/static_cast<double>(n_r); 
  const double s_l = p_l / pow(rho_l, Gamma); 
  const double s_r = p_r / pow(rho_r, Gamma); 
  
  for (p_sz i = 0; i != n_l; ++i) {
    ParticleData particle; 
    particle.PutHsml(2 * dx_l); 
    particle.PutPos(-5.e-1 + i * dx_l, 0);  
    particle.PutVel(v_l, 0); 
    particle.PutMass(mgas); 
    particle.PutEntropy(s_l); 
    particle.PutDensity(rho_l); 
    sph.push_back(particle); 
  }

  for (p_sz i = 0; i != n_r; ++i) {
    ParticleData particle;   
    particle.PutHsml(2 * dx_r); 
    particle.PutPos(i * dx_r, 0);  
    particle.PutVel(v_r, 0); 
    particle.PutMass(mgas); 
    particle.PutEntropy(s_r); 
    particle.PutDensity(rho_r); 
    sph.push_back(particle); 
  }

}
#endif

#ifdef TWO_D_KH
void init_kh(vector<ParticleData>& sph) {

  typedef vector<ParticleData>::size_type p_sz; 


  const real rho_h = 2.0; 
  const real rho_l = 1.0; 
  const real p_h = 2.5; 
  const real p_l = 2.5; 
  const real v_h = -0.5; 
  const real v_l = 0.5;
  const double s_h = p_h / pow(rho_h, Gamma); 
  const double s_l = p_l / pow(rho_l, Gamma); 

  const int ngrid_h = Ngrid; 
  const int ngrid_l = sqrt(sq(ngrid_h) *rho_l/rho_h); 

  const p_sz n_h = 0.5 * sq(ngrid_h); 
  const p_sz n_l = 0.5 * sq(ngrid_l); 

  const real mgas = 0.5 * (rho_h + rho_l) * (BoxSize) / static_cast<double>(n_h + n_l); 
  cout << "ngrid_h = " << ngrid_h << " ngrid_l = " << ngrid_l << endl; 
  cout << "npart_h = " << n_h << " npart_l = " << n_l << " ntot = " << n_l + n_h << endl; 

  sph.resize(n_l + n_h); 
  sph.clear(); 

  int half_grid = 0.5 * ngrid_h; 
  for (int i = 0; i != ngrid_h; ++i) {
    for (int j = 0; j != half_grid; ++j) {
      ParticleData particle; 
      double dx = sqrt(0.5 * sq(BoxSize)/static_cast<double>(n_h)); 
      particle.PutHsml(2*dx); 
      double x = (0.5 + i) * dx; 
      particle.PutPos(x, 0); 
      double y = 0.25 + (0.5 + j) * dx; 
      particle.PutPos(y, 1); 

      particle.PutVel(v_h, 0); 
      double vy = 0; 
      if (fabs(particle.GetPos(1) - 0.5 - 0.25) < 0.025) 
        vy = 0.025 * sin(-2 * M_PI * (particle.GetPos(0))*6); 
      else if (fabs(particle.GetPos(1) - 0.5 + 0.25) < 0.025) 
        vy = 0.025 * sin(2 * M_PI * (particle.GetPos(0))*6); 
      particle.PutVel(vy, 1); 
      particle.PutMass(mgas); 
      particle.PutEntropy(s_h); 
      particle.PutDensity(rho_h); 
      sph.push_back(particle); 
    }
  }

  half_grid = 0.5 * ngrid_l; 
  for (int i = 0; i != ngrid_l; ++i) {
    for (int j = 0; j != half_grid; ++j){
      ParticleData particle; 
      double dx = sqrt(0.5 * sq(BoxSize)/static_cast<double>(n_l)); 
      particle.PutHsml(2*dx); 
      double x = (0.5 + i) * dx; 
      particle.PutPos(x, 0); 
      double y = (0.5 + j) * dx; 
      if (y > 0.25) 
        y += 0.5; 
      particle.PutPos(y, 1); 
      particle.PutVel(v_l, 0); 
      double vy = 0; 
      if (fabs(particle.GetPos(1) - 0.5 - 0.25) < 0.025) 
        vy = 0.025 * sin(-2 * M_PI * (particle.GetPos(0))*6); 
      else if (fabs(particle.GetPos(1) - 0.5 + 0.25) < 0.025) 
        vy = 0.025 * sin(2 * M_PI * (particle.GetPos(0))*6); 
      particle.PutVel(vy, 1); 
      particle.PutMass(mgas); 
      particle.PutEntropy(s_l); 
      particle.PutDensity(rho_l); 
      sph.push_back(particle); 
    }
  }

  cout << "sph.size() = " << sph.size() << endl; 
}
#endif

#ifdef THREE_D_SEDOV
void init_sedov(vector<ParticleData>& sph) 
{
  const double Mtot = 1; 
  const double mass = Mtot/static_cast<double>(cube(Ngrid)); 
  const double dx = 1.e0/static_cast<double>(Ngrid); 
  const double density = Mtot/(cube(BoxSize)); 
  const int Nby2 = Ngrid/2; 
  const double E_inject = 1; 
  double U_inject = E_inject/mass; 
  double wtot = 0; 
  for (int i = 0; i != Ngrid; ++i) {
    for (int j = 0; j != Ngrid; ++j) {
      for (int k = 0; k != Ngrid; ++k) {
        ParticleData p; 
        double box_half = 0.5 * BoxSize; 
        real x = (i + 0.5) * dx; 
        real y = (j + 0.5) * dx; 
        real z = (k + 0.5) * dx; 
        p.PutPos(x, 0); 
        p.PutPos(y, 1); 
        p.PutPos(z, 2); 

        p.PutMass(mass); 
        p.PutHsml(2*dx); 
        p.PutDensity(density); 

        
        double h = 2. * dx; 
        double hinv = 1.e0/h; 
        double hinv3 = cube(hinv); 

        double r = sqrt( sq(x - box_half) + sq(y - box_half) + sq(z - box_half) ); 
        double u = r * hinv; 
        if (u < 5.e-1) {
          double wk = 1. + 6. * (u - 1) * sq(u); 
          wk *= KernelCoeff * hinv3; 
          wtot += wk; 
        } else if (u < 1.) {
          double wk = 2. * cube(1. - u); 
          wk *= KernelCoeff * hinv3; 
          wtot += wk; 
        } 
#if 0
        if (i == Nby2 && j == Nby2 && k == Nby2) 
          p.PutEntropy(Gamma_1 * U_inject / pow(density, Gamma_1)); 
        else 
          p.PutEntropy(1.e-10 * Gamma_1 * U_inject / pow(density, Gamma_1)); 
#endif 

        sph.push_back(p); 
      }
    }
  }

  for (vector<ParticleData>::size_type i = 0; i != sph.size(); ++i) { 
    double box_half = 0.5 * BoxSize; 
    real x = sph[i].GetPos(0) - box_half; 
    real y = sph[i].GetPos(1) - box_half; 
    real z = sph[i].GetPos(2) - box_half; 

    double h = 2. * dx; 
    double hinv = 1.e0/h; 
    double hinv3 = cube(hinv); 

    double r = sqrt( x*x + y*y + z*z ); 
    double u = r * hinv; 
    
    if (u < 5.e-1) {
      double wk = 1. + 6. * (u - 1) * sq(u); 
      wk *= KernelCoeff * hinv3; 
      sph[i].PutEntropy(wk *Gamma_1 * U_inject / pow(density, Gamma_1) / wtot); 
    } else if (u < 1.) {
      double wk = 2. * cube(1. - u); 
      wk *= KernelCoeff * hinv3; 
      sph[i].PutEntropy(wk *Gamma_1 * U_inject / pow(density, Gamma_1) / wtot); 
    } else {
      sph[i].PutEntropy(1.e-10*Gamma_1 * U_inject / pow(density, Gamma_1)); 
    }

  }
}
#endif
