#ifndef PARTICLEDATA_H
#define PARTICLEDATA_H
//-----------------------
#include "config.h"
#include <iostream>
#include <vector>
#include <algorithm>
#include <stdexcept>

#if defined(ONE_D_SHOCKTUBE) || defined(TWO_D_KH) || defined(THREE_D_SEDOV) 
extern int Ngrid;
#endif 
//------ Global variables for SPH calculation ------------
extern double Gamma,  Gamma_1; 
extern double ArtBulkViscConst, CourantFac; 
#ifdef REDUCEVISC
extern double MinAlpha, ViscParam, ViscDecayLength; 
#endif
#ifdef AC
extern double ArtBulkCondConst; 
#endif
extern double Nngb; // smoothed number of neighbours
extern double MinNngb, MaxNngb;  // boundaries 
//-------------------------------------------------------


class ParticleData {
  public:
    ParticleData(void); 

    void PutHsml(const double& h) { hsml = h; }
    double GetHsml() const { return hsml; }

    void PutPos(const real& x, const int& dim) { pos[dim]  = x; }
    real GetPos(const int& dim) const { return pos[dim]; }
    real& Pos(const int& dim) { return pos[dim]; }
    real* Pos() { return pos; }

    void PutVel(const real& v, const int& dim) { vel[dim] = v; }
    real GetVel(const int& dim) const { return vel[dim]; }
    
    void PutDensity(const real& rho) { density = rho; }
    real GetDensity() const { return density; }

    void PutMass(const real& m) { mass = m; }
    real GetMass() const { return mass; }

        void PutDivVel(const real& v) { divv = v; }
    real GetDivVel() const { return divv; }

#if (DIM > 1)
    void PutCurl(const real& c) { curl = c; }
    real GetCurl() const { return curl; }
#endif

    void PutDhsmlDensityFactor(const real& d) {
      dhsmldensityfactor = d; 
    }

    real GetDhsmlDensityFactor() const {
      return dhsmldensityfactor; 
    }

#ifdef USE_SMOOTHED_P
    void PutSmoothedPressure(const real& p) { smoothed_presure = p; }
    real GetSmoothedPressure() const { return smoothed_presure; }
#endif
#ifdef REDUCEVISC
    void PutAlpha(const real& a) { alpha = a; }
    real GetAlpha() const { return alpha; }; 

    void PutDtAlpha(const real& da) { dt_alpha = da; }
    real GetDtAlpha() const { return dt_alpha; }
#endif 
    
#ifdef AC
    void PutAlphaU(const real& a) { alpha_u = a; }
    real GetAlphaU() const { return alpha_u; }

    void PutDtAlphaU(const real& da) { dt_alpha_u = da; }
    real GetDtAlphaU() const { return dt_alpha_u; }
#endif

    void PutAcc(const real& a, const int& dim) { acc[dim] = a; }
    real GetAcc(const int& dim) { return acc[dim]; }
    
    void PutEntropy(const real& s) { entropy = s; }
    real GetEntropy() const { return entropy; }

    void PutDtEntropy(const real& ds) { dt_entropy = ds; }
    real GetDtEntropy() const { return dt_entropy; }

    void PutDtEntropyOld(const real& ds) { dt_entropy_old = ds; }
    real GetDtEntropyOld() const { return dt_entropy_old; }

    void PutMaxSigVel(const real& sv) { max_signal_vel = sv; }
    real GetMaxSigVel() const { return max_signal_vel; }


    void UpdatePos(const double dt) {
      for (int dim = 0; dim != DIM; ++dim) 
        pos[dim] += half_step_v[dim]*dt; 
    }; 

    void UpdateVel(const double dt) {
      for (int dim = 0; dim != DIM; ++dim) 
        vel[dim] = half_step_v[dim] + 5.e-1 * acc[dim] * dt; 

    }

    void FirstKick(const double dt) {
      for (int dim = 0; dim != DIM; ++dim) 
        half_step_v[dim] = vel[dim] + acc[dim] * dt; 
    }

    void Kick(const double dt) {
      for (int dim = 0; dim != DIM; ++dim) 
        half_step_v[dim] += acc[dim] * dt; 
    }

    void UpdateEntropy(const double dt) { 
      entropy += 0.5 * (dt_entropy_old + dt_entropy) * dt; 
      if (entropy <= 0) {
        std::cerr << "Entropy = " << entropy << " DtDs = " << dt_entropy << std::endl; 
        throw std::domain_error("negative entropy!!"); 
      }
    }

    double CalcDt() const { return 2 * CourantFac * hsml / max_signal_vel; }

#ifdef REDUCEVISC
    void UpdateAlpha(const double dt) {
      alpha += dt_alpha * dt; 
      alpha = std::min(alpha, static_cast<real>(ArtBulkViscConst)); 
      alpha = std::max(alpha, static_cast<real>(MinAlpha)); 
    }
#endif

#ifdef AC
    void UpdateAlphaU(const double dt) {
      alpha_u += dt_alpha_u * dt; 
      alpha_u = std::max(static_cast<real>(0), alpha_u); 
      alpha_u = std::min(alpha_u, static_cast<real>(ArtBulkViscConst)); 
    }
#endif

  private: 
    real pos[DIM]; 
    real vel[DIM]; 
    real half_step_v[DIM]; 
    real acc[DIM]; 
    real mass; 
    real entropy; 
    real dt_entropy; 
    real dt_entropy_old; 
    real density; 
#ifdef USE_SMOOTHED_P
    real smoothed_presure; 
#endif
    double hsml; 
    real divv; 
#if (DIM > 1)
    real curl; 
#endif
    real dhsmldensityfactor; 
    real max_signal_vel; 
#ifdef REDUCEVISC
    real alpha; 
    real dt_alpha; 
#endif
#ifdef AC
    real alpha_u; 
    real dt_alpha_u; 
#endif
}; 

//-----------------------
#endif
