#include "config.h"
#include "run.h"
#include "particledata.h"
#include "tree.h"
#include "sph.h"
#include "output.h"
#include <iostream>
#include <vector>
#include <algorithm>

using std::vector; 
using std::min; 
using std::cout; 
using std::endl; 

double calc_dt (vector<ParticleData>& sph, const double& time, const int snap_num); 
void first_kick(const double dt, vector<ParticleData>& sph); 
void kick(const double dt, vector<ParticleData>& sph); 
void drift(const double dt, vector<ParticleData>& sph);  

void run(double dt, vector<ParticleData>& sph) 
{
  vector<ParticleList> plist; 
  vector<NodeInfo> tree; 
  double time = BegTime; 
  int snap_num = 0; 
  do {

    if (dt == 0) {
      cout << "start building tree" << endl; 
      BuildTree(tree, plist, sph);  
      cout << "start calculating density" << endl; 
      CalcDensity(tree, plist, sph); 
#ifdef USE_SMOOTHED_P
      CalcSmoothedPresure(tree, plist, sph); 
#endif
      cout << "start updating hmax" << endl; 
      UpdateHmax(tree, plist, sph); 
      cout << "start hydroforce calculation" << endl; 
      CalcHydroForce(tree, plist, sph); 
      cout << "done!" << endl; 
      Output(sph, snap_num); 
      dt = calc_dt(sph, time, snap_num); 
      first_kick(dt * 5.e-1, sph); 
    } else {
      double dt_old = dt; 
      dt = calc_dt(sph, time, snap_num); 
      kick(5.e-1 * (dt + dt_old), sph); 
    }
    drift(dt, sph); 
    BuildTree(tree, plist, sph); 
    CalcDensity(tree, plist, sph); 
#ifdef USE_SMOOTHED_P
      CalcSmoothedPresure(tree, plist, sph); 
#endif
    UpdateHmax(tree, plist, sph); 
    CalcHydroForce(tree, plist, sph); 
    time += dt; 
    cout << "Time = " << time << endl; 
    if (time >= (snap_num + 1) * DtOutput) {
      cout << "Output" << endl; 
      Output(sph, snap_num+1); 
      snap_num ++; 
    }
  } while (time < MaxTime); 
}

double calc_dt (vector<ParticleData>& sph, const double& time, const int snap_num)
{

  double dt = BegTime + DtOutput * (snap_num + 1) - time; 
  typedef vector<ParticleData>::size_type p_sz; 
  for (p_sz i = 0; i != sph.size(); ++i) 
    dt = min(dt, sph[i].CalcDt()); 

  return dt; 
}

void first_kick(const double dt, vector<ParticleData>& sph) 
{
  typedef vector<ParticleData>::size_type p_sz; 
  for (p_sz i = 0; i < sph.size(); ++i)
    sph[i].FirstKick(dt); 
}

void kick(const double dt, vector<ParticleData>& sph) 
{
  typedef vector<ParticleData>::size_type p_sz; 
  for (p_sz i = 0; i < sph.size(); ++i)
    sph[i].Kick(dt); 
}

void drift(const double dt, vector<ParticleData>& sph) 
{
  typedef vector<ParticleData>::size_type p_sz; 

  for (p_sz i = 0; i < sph.size(); i++) { 
    sph[i].UpdatePos(dt); 
    for (int dim = 0; dim < DIM; dim++) {
#ifdef PERIODIC
      if (sph[i].GetPos(dim) < 0) 
        sph[i].PutPos(sph[i].GetPos(dim) + BoxSize, dim); 
      else if (sph[i].GetPos(dim) >= BoxSize) 
        sph[i].PutPos(sph[i].GetPos(dim) - BoxSize, dim); 
#endif
    }
    sph[i].UpdateVel(dt); 
    sph[i].UpdateEntropy(dt); 
    sph[i].PutDtEntropyOld(sph[i].GetDtEntropy()); 
#ifdef REDUCEVISC
    sph[i].UpdateAlpha(dt); 
#endif
#ifdef AC
    sph[i].UpdateAlphaU(dt); 
#endif
  }
}



