#include "config.h"
#include "utils.h"
#include "particledata.h"
#include "tree.h"
#include "neighbour.h"
#include "sph.h"

#include <iostream>
#include <vector>
#include <stdexcept>
#include <algorithm>
#include <stdexcept>
#include <cmath>

#ifdef _OPENMP
#include <omp.h>
#endif 

using std::cout; 
using std::endl; 
using std::vector; 
using std::max; 
using std::min; 
using std::domain_error; 

typedef vector<ParticleData>::size_type p_sz; 


//---------------------------------------------------------------------------
void calc_density_i(const vector<NodeInfo>& tree, const vector<ParticleList>& plist, 
    vector<ParticleData>& sph, p_sz i); 
bool change_hsml(double& hsml, double& left, double& right, const real& smoothed_num); 
void calc_hydro_force_i(const vector<NodeInfo>& tree, const vector<ParticleList>& plist, 
    vector<ParticleData>& sph, const p_sz i); 
#ifdef USE_SMOOTHED_P
void calc_smoothed_pressure_i(const vector<NodeInfo>& tree, const vector<ParticleList>& plist, vector<ParticleData>& sph, const p_sz i); 
#endif 
//---------------------------------------------------------------------------

#ifndef _OPENMP
vector<p_sz> nlist; // just to avoid overhead of construncting and destructiong.... 
#endif

void CalcDensity(const vector<NodeInfo>& tree, const vector<ParticleList>& plist, vector<ParticleData>& sph)
{
#ifndef FUJITSU
  const p_sz loop_size = sph.size(); 
#else 
  const int loop_size = sph.size(); 
#endif
#pragma omp parallel for 
#ifndef FUJITSU
  for (p_sz i = 0; i < loop_size; ++i) 
#else 
  for (int i = 0; i < loop_size; ++i) 
#endif 
  {
    calc_density_i(tree, plist, sph, i); 
  }
}


void calc_density_i(const vector<NodeInfo>& tree, const vector<ParticleList>& plist, vector<ParticleData>& sph, p_sz i)
{
  double left = 0; 
  double right = 0; 
  real smoothed_num = 0;  
  double h = sph[i].GetHsml(); 

#ifdef _OPENMP
  vector<p_sz> nlist; 
#endif

  while (smoothed_num < MinNngb || smoothed_num > MaxNngb) {

    double h2 = h * h; 
    double hinv = 1.e0/h; 
    double h3 = H3(h); 
    double hinv3 = 1.e0/h3; 
    double hinv4 = hinv * hinv3; 
    double rho = 0; 
    double divvel = 0; 
#if (DIM == 2) 
    double curl = 0; 
#elif (DIM == 3) 
    double rot[DIM]; 
    for (int dim = 0; dim != DIM; ++dim)
      rot[dim] = 0; 
#endif

    double dhsmldensity = 0; 

    FindNeighbours(nlist, tree, plist, sph, sph[i].Pos(), h); 
    smoothed_num = 0; 
    double dx[DIM]; 
    for (vector<p_sz>::size_type n = 0; n != nlist.size(); ++n) {
      p_sz j = nlist[n]; 
      for (int dim = 0; dim != DIM; ++dim) {
        dx[dim] = sph[i].GetPos(dim) - sph[j].GetPos(dim); 
#ifdef PERIODIC
        periodic_dx(dx[dim]); 
#endif
      }
      double r2 = 0; 
      for (int dim = 0; dim != DIM; ++dim)
        r2 += sq(dx[dim]); 

      if (r2 < h2) {
        double r = sqrt(r2); 
        double u = r * hinv; 
        double wk, dwk; 
        if (u < 5.e-1) {
          wk = 1. + 6. * (u - 1) * sq(u); 
          dwk = -12. * u + 18. * sq(u); 
        } else {
          wk = 2. * cube(1. - u); 
          dwk = -6. * sq(1. - u); 
        }
        wk *= KernelCoeff * hinv3; 
        dwk *= KernelCoeff * hinv4; 
        smoothed_num += NormCoeff * h3 * wk; 

        real m_j = sph[j].GetMass(); 

        rho += m_j * wk; 

        dhsmldensity += - m_j * (DIM*hinv*wk + u*dwk);  // drho/dh

        if (r > 0) {
          double fac = m_j * dwk / r;
          real dv[DIM]; 
          for (int dim = 0; dim != DIM; ++dim)
            dv[dim] = sph[i].GetVel(dim) - sph[j].GetVel(dim); 

          for (int dim = 0; dim != DIM; ++dim) {
            divvel -= fac * dx[dim] * dv[dim]; 
#if (DIM == 2)
            curl += fac * (dx[0]*dv[1] + dx[1]*dv[0]); 
#elif (DIM == 3)
            int dim1 = dim + 1; 
            int dim2 = dim + 2; 
            if (dim1 >= DIM) 
              dim1 -= DIM; 
            if (dim2 >= DIM)
              dim2 -= DIM; 
            rot[dim] += fac * (dx[dim2]*dv[dim1] - dx[dim1]*dv[dim2]); 
#endif
          } 
        }
      }
    }

    bool break_the_loop = change_hsml(h, left, right, smoothed_num); 

    if (break_the_loop) {
      // Ok now we update density of this particle
      sph[i].PutDensity(rho); 
      sph[i].PutHsml(h); 
      sph[i].PutDivVel(divvel/rho); 
      sph[i].PutDhsmlDensityFactor(
          1.e0/(1 + h*dhsmldensity/(DIM*rho))
          ); 
#if (DIM == 2)
      sph[i].PutCurl(fabs(curl)/rho); 
#elif (DIM == 3) 
      sph[i].PutCurl(sqrt(sq(rot[0]) + sq(rot[1]) + sq(rot[2]))/rho); 
#endif
#ifdef REDUCEVISC
      double pressure = sph[i].GetEntropy() * pow(rho, Gamma); 
      double cs_h = sqrt(Gamma * pressure / rho) / h; 
      double dt_alpha = -(sph[i].GetAlpha() - MinAlpha) * ViscParam * cs_h 
        + (ArtBulkViscConst - sph[i].GetAlpha()) * max(0.0, -static_cast<double>(sph[i].GetDivVel())); 
      sph[i].PutDtAlpha(dt_alpha);
#endif
      break; 
    }
  } // end of a long while loop 
}

bool change_hsml(double& hsml, double& left, double& right, const real& smoothed_num)
{
  if (smoothed_num >= MinNngb && smoothed_num <= MaxNngb) 
    return true; 

  if (smoothed_num < MinNngb) {
    left = max(hsml, left); 
  } else if (smoothed_num > MaxNngb) {
    if (right > 0) 
      right = min(hsml, right); 
    else 
      right = hsml; 
  }

  real rdim = static_cast<real>(DIM); 

  if (left > 0 && right > 0) {
    if (right - left < 1.e-3 * left) // This one should be ok
      return true; 
#if (DIM == 1) 
    hsml = 0.5 * (left + right); 
#else 
    hsml = pow(0.5 * (pow(left, rdim) + pow(right, rdim)), 1./rdim); 
#endif
  } else {
    if (right == 0 && left == 0) 
      throw domain_error("Cannot occur!"); 

    if (right == 0) { // left > 0
      hsml *= 1.2; 
    } else { // right > 0
      hsml *= 0.8; 
    }
  }
  return false; 
}

void CalcHydroForce(const vector<NodeInfo>& tree, const vector<ParticleList>& plist, vector<ParticleData>& sph)
{
#ifndef FUJITSU
  const p_sz loop_size = sph.size(); 
#else 
  const int loop_size = sph.size(); 
#endif
#pragma omp parallel
#ifndef FUJITSU
  for (p_sz i = 0; i < loop_size; ++i) 
#else 
  for (int i = 0; i < loop_size; ++i) 
#endif
  {
    // initialization 
    for (int dim = 0; dim != DIM; ++dim) 
      sph[i].PutAcc(0, dim); 
    sph[i].PutDtEntropy(0); 
    sph[i].PutMaxSigVel(0); 
    calc_hydro_force_i(tree, plist, sph, i); 
  }

#ifdef ONE_D_SHOCKTUBE // impose boundary conditions
  for (p_sz i = 0; i < 4; ++i) {
    for (int dim = 0; dim != DIM; ++dim) 
      sph[i].PutAcc(0, dim); 
    sph[i].PutDtEntropy(0); 
  }

  for (p_sz i = sph.size() - 1; i > sph.size() - 5; --i) {
    for (int dim = 0; dim != DIM; ++dim) 
      sph[i].PutAcc(0, dim); 
    sph[i].PutDtEntropy(0); 
  }
#endif
}

void calc_hydro_force_i(const vector<NodeInfo>& tree, const vector<ParticleList>& plist, 
    vector<ParticleData>& sph, const p_sz i)
{
  const real rho_i = sph[i].GetDensity(); 
  const real h_i = sph[i].GetHsml(); 
#ifndef USE_SMOOTHED_P
  const real pressure_i = sph[i].GetEntropy() * pow(rho_i, Gamma); 
#else 
  const real pressure_i = sph[i].GetSmoothedPressure();  
#endif

  const double soundspeed_i = sqrt(Gamma * pressure_i/rho_i); 

#ifdef _OPENMP
  vector<p_sz> nlist; 
#endif

#if (DIM > 1) 
  const double f1 = fabs(static_cast<double>(sph[i].GetDivVel()))/(
      fabs(static_cast<double>(sph[i].GetDivVel()))  + sph[i].GetCurl() + Epsilon * soundspeed_i/h_i
      ); 
#else 
  const double f1 = 1; 
#endif

  const double p_over_rho2_i = pressure_i / sq(rho_i) * sph[i].GetDhsmlDensityFactor(); 
#ifdef AC
  const double u_i = sph[i].GetEntropy() * pow(rho_i, Gamma_1)/Gamma_1; 
#endif
  //const double p_over_rho2_i = pressure_i / sq(rho_i);  
  const double h_i2 = sq(h_i); 

  double dv[DIM], dx[DIM]; 
  double dwk_i, dwk_j, u; 

  FindNeighbourPairs(nlist, tree, plist, sph, sph[i].Pos(), h_i); 

  double acc[DIM]; 
  for (int dim = 0; dim != DIM; ++dim)
    acc[dim] = 0; 

  double dt_entropy = 0; 

#ifdef AC
  double ddu = 0; 
#endif

  for (vector<p_sz>::size_type n = 0; n < nlist.size(); ++n) { // j loop 
    p_sz j = nlist[n]; 
    for (int dim = 0; dim != DIM; ++dim) {
      dx[dim] = sph[i].GetPos(dim) - sph[j].GetPos(dim); 
#ifdef PERIODIC
      periodic_dx(dx[dim]); 
#endif
    }

    double r2 = 0; 
    for (int dim = 0; dim != DIM; ++dim)
      r2 += sq(dx[dim]); 

    double h_j = sph[j].GetHsml(); 
    double h_j2 = sq(h_j); 

    if (r2 < h_i2 || r2 < h_j2) {
      double r = sqrt(r2); 
      if (r > 0) {
        real rho_j = sph[j].GetDensity(); 
#ifndef USE_SMOOTHED_P
        real pressure_j = sph[j].GetEntropy() * pow(rho_j, Gamma); 
#else 
        real pressure_j = sph[j].GetSmoothedPressure();  
#endif

        double soundspeed_j = sqrt(Gamma * pressure_j / rho_j); 
        double p_over_rho2_j = pressure_j / sq(rho_j) * sph[j].GetDhsmlDensityFactor(); 
        //double p_over_rho2_j = pressure_j / sq(rho_j); 
#ifdef AC
        double u_j = sph[j].GetEntropy() * pow(rho_j, Gamma_1)/Gamma_1;
#endif
        for (int dim = 0; dim != DIM; ++dim) 
          dv[dim] = sph[i].GetVel(dim) - sph[j].GetVel(dim); 
        double vdotr = 0; 
        for (int dim = 0; dim != DIM; ++dim) 
          vdotr += dx[dim] * dv[dim]; 

        if (r2 < h_i2) {
          double hinv = 1.e0/h_i; 
          double hinv3 = H3(hinv); // this function takes care of the dimension
          double hinv4 = hinv *hinv3; 
          u = r * hinv; 
          if (u < 5.e-1) 
            dwk_i = KernelCoeff * hinv4 * (-12*u + 18*sq(u)); 
          else 
            dwk_i = - KernelCoeff * hinv4 * 6 * sq(1.e0 - u); 
        } else {
          dwk_i = 0; 
        }

        if (r2 < h_j2) {
          double hinv = 1.0/h_j; 
          double hinv3 = H3(hinv); 
          double hinv4 = hinv3 * hinv; 
          u = r * hinv; 
          if (u < 5.e-1) 
            dwk_j = KernelCoeff * hinv4 * (-12*u + 18*sq(u)); 
          else 
            dwk_j = - KernelCoeff * hinv4 * 6 * sq(1.e0 - u); 
        } else {
          dwk_j = 0; 
        }

        // now we have dwk_i and dwk_j
        sph[i].PutMaxSigVel( 
            max(static_cast<double>(sph[i].GetMaxSigVel()), soundspeed_i + soundspeed_j) 
            ); 

        double visc; 
#ifdef AC
        double dt_entropy_ac = 0; 
#endif
        // things for the artificial viscosity 
        if (vdotr < 0) { // two particles are approaching!!
          double rho_ij = 0.5 * (rho_i + rho_j); 
          double mu_ij = vdotr / r; 
          //double vsig = (soundspeed_i + soundspeed_j - 3 * mu_ij); 
          double vsig = 0.5 * (soundspeed_i + soundspeed_j - 2 * mu_ij); 
          sph[i].PutMaxSigVel(
              max(static_cast<double>(sph[i].GetMaxSigVel()), vsig)
              ); 
#if (DIM > 1)
          double f2 = fabs(static_cast<double>(sph[j].GetDivVel())) / ( 
                fabs(static_cast<double>(sph[j].GetDivVel()))  + sph[j].GetCurl() + Epsilon * soundspeed_j/h_j
                ); 
#else 
          double f2 = 1; 
#endif 

#ifndef REDUCEVISC
          visc = 0.5 * ArtBulkViscConst * vsig * (-mu_ij) / rho_ij * (f1 + f2); 
#else // else ndef REDUCEVISC 
          double BulkVisc_ij = 0.5 * (sph[i].GetAlpha() + sph[j].GetAlpha()); 
          visc = 0.5 * BulkVisc_ij * vsig * (-mu_ij) / rho_ij * (f1 + f2); 
#endif // end of REDUCEVISC

#ifdef AC 
          // things for the artificial conductivity 
          ddu += 2 * sph[j].GetMass() * (u_i - u_j) * 0.5 * fabs(dwk_i + dwk_j)/(r * rho_j); 
          double vsig_u = sqrt(fabs(pressure_i - pressure_j)/rho_ij); 
          dt_entropy_ac = sph[j].GetMass() * 0.5 * (sph[i].GetAlphaU() + sph[j].GetAlphaU()) * vsig_u * (u_i - u_j) * 
            0.5 * (sph[i].GetDhsmlDensityFactor() * dwk_i + sph[j].GetDhsmlDensityFactor()*dwk_j)/rho_ij;
#endif 

        } else {
          visc = 0; 
#ifdef AC
          /*
          double rho_ij = 0.5 * (rho_i + rho_j); 
          ddu += 2 * sph[j].GetMass() * (u_i - u_j) * 0.5 * fabs(dwk_i + dwk_j)/(r * rho_j); 
          double vsig_u = sqrt(fabs(pressure_i - pressure_j)/rho_ij); 
          dt_entropy_ac = sph[j].GetMass() * 0.5 * (sph[i].GetAlphaU() + sph[j].GetAlphaU()) * vsig_u * (u_i - u_j) * 
            0.5 * (sph[i].GetDhsmlDensityFactor() * dwk_i + sph[j].GetDhsmlDensityFactor()*dwk_j)/rho_ij;
            */
          dt_entropy_ac = 0; 
#endif
        }


        double hfc_visc = 0.5 * sph[j].GetMass() * visc 
         * (sph[i].GetDhsmlDensityFactor() * dwk_i + sph[j].GetDhsmlDensityFactor()*dwk_j) / r; 
         //* (dwk_i + dwk_j) / r; 
        double hfc = hfc_visc + sph[j].GetMass() *  (p_over_rho2_i*dwk_i + p_over_rho2_j*dwk_j) / r; 
        for (int dim = 0; dim != DIM; ++dim)
           acc[dim] -= hfc * dx[dim]; 
        dt_entropy += 0.5 * hfc_visc * vdotr; 

#ifdef AC 
        dt_entropy += dt_entropy_ac; 
#endif
      }
    }
  }

  for (int dim = 0; dim != DIM; ++dim)
    sph[i].PutAcc(acc[dim], dim); 
  dt_entropy *=  Gamma_1 / pow(rho_i, Gamma_1); // internal energy to entropy 
  sph[i].PutDtEntropy(dt_entropy); 
#ifdef AC
  double cs_h = sqrt(Gamma * pressure_i / rho_i) / h_i; 
  double dt_alpha_u = - sph[i].GetAlphaU() * ViscParam * cs_h  + ArtBulkCondConst * h_i * fabs(ddu) / sqrt(u_i + Epsilon); 
  sph[i].PutDtAlphaU(dt_alpha_u); 
#endif
}

#ifdef USE_SMOOTHED_P
void CalcSmoothedPresure(const vector<NodeInfo>& tree, const vector<ParticleList>& plist, vector<ParticleData>& sph) 
{
  for (p_sz i = 0; i < sph.size(); ++i) {
    calc_smoothed_pressure_i(tree, plist, sph, i); 
  }
}

void calc_smoothed_pressure_i(const vector<NodeInfo>& tree, const vector<ParticleList>& plist, vector<ParticleData>& sph, const p_sz i) 
{
  double smoothed_pressure = 0; 

  double dx[DIM]; 
  double h = sph[i].GetHsml(); 
  double h2 = sq(h); 
  double hinv = 1.e0/h; 
  double hinv3 = H3(hinv); // take care of dimension

#ifdef _OPENMP
  vector<p_sz> nlist; 
#endif

  FindNeighbours(nlist, tree, plist, sph, sph[i].Pos(), h); 
  for (vector<p_sz>::size_type n = 0; n != nlist.size(); ++n) {
    p_sz j = nlist[n]; 
    for (int dim = 0; dim != DIM; ++dim) { 
      dx[dim] = sph[i].GetPos(dim) - sph[j].GetPos(dim); 
#ifdef PERIODIC
      periodic_dx(dx[dim]); 
#endif
    }
    double r2 = 0; 
    for (int dim = 0; dim != DIM; ++dim)
      r2 += sq(dx[dim]); 
    if (r2 < sq(h)) {
      double r = sqrt(r2); 
      double u = r * hinv; 
      double wk; 
      if (u < 5.e-1) {
        wk = 1. + 6. * (u - 1) * sq(u); 
      } else {
        wk = 2. * cube(1. - u); 
      }
      wk *= KernelCoeff * hinv3; 
      smoothed_pressure += sph[j].GetMass() * sph[j].GetEntropy() * pow(sph[j].GetDensity(), Gamma_1) * wk; 
    }
  }
  sph[i].PutSmoothedPressure(smoothed_pressure); 
}
#endif

